import json
import math
from random import random
from typing import List
import Adafruit_IO as AIO
from dotenv import dotenv_values

aio = None # Attempted to be initialized with .env credentials

defaultSchema = {
  "group": "brad",
  "ping_pi": "ping", # Name is pi_ping, key is 'ping'
  "ping_host": "host-ping",
  "test": "test",
  
  "pi_status": "pi-status",
  "host_status": "host-status",

  "data": "test-data",
  "control": "test-control",
  "stream": "ir-stream",
}

class Aio:
  client: AIO.Client = None
  schema: dict[str, str] = {}
  
  def __init__(self, username: str, key: str, scheme: dict[str, str] = defaultSchema):
    self.client = AIO.Client(username, key)
    self.schema = scheme
  
  # @classmethod
  def send(cls, group: str, feed: str, *, data):
      fullName = group + "." + feed
      cls.client.send(fullName, data)
  
  # @classmethod
  def receive(cls, group: str, feed: str):
      fullName = group + "." + feed
      print(f"Getting from {fullName=}")
      return cls.client.receive(fullName).value

  # @classmethod
  def send_schema(cls, scheme_option, data):
    cls.send(cls.schema["group"], cls.schema[scheme_option], data=data)
    
  # @classmethod
  def receive_schema(cls, scheme_option):
    return cls.receive(cls.schema["group"], cls.schema[scheme_option])

  def send_data(self, data):
      self.send_schema("data", data=data)
  
  def pi_status_send(self, data):
    self.send_schema("pi_status", data=data)
  
  def _get_pi_status(self):
    return self.receive_schema("pi_status")
  
  def host_status_send(self, data):
    self.send_schema("host_status", data=data)
  
  def _get_host_status(self):
    return self.receive_schema("host_status")
      
  def pi_ping(self, *, streaming: bool = True, inactive: bool = False):
    code: int = 0
    if not inactive: 1 + int(streaming)
    else: code = -1
    self.send_schema("ping_pi", code)
  
  def _get_pi_ping_status(self):
    return self.receive_schema("ping_pi")
    
  def host_ping(self, *, receiving: bool = True, inactive: bool = False):
    code: int = 0
    if not inactive: 1 + int(receiving)
    else: code = -1
    self.send_schema("ping_host", code)
  
  def _get_host_ping_status(self):
    return self.receive_schema("ping_host")  

  def receive_control(self):
      return self.receive_schema("control")
  
  def send_stream(self, data):
      self.send_schema("stream", data=data)
  
  def receive_stream(self):
      return self.receive_schema("stream")
  
  def send_stream_data(self, data):
      self.send_stream(json.dumps({
        "stream": data,
      }))
  
  def receive_stream_data(self) -> List[int] | str:
    return json.loads(self.receive_stream())["stream"]
  
def reset_aio_stream(stratagy: str = "lines", *, aio=aio):
  fakeStream = []
  for y in range(24):
    for x in range(32):
      if stratagy == "waves":
        fakeStream.append(x+y)
      elif stratagy == "lines":
        fakeStream.append(x)
      elif stratagy == "columns":
        fakeStream.append(y)
      elif stratagy == "random" or True:
        fakeStream.append(int(random()+1)*69)
  aio.send_stream_data(fakeStream)
  stream_data = aio.receive_stream_data()
  assert len(stream_data) == 32*24

env_variables = dotenv_values()
try:
  defaultPassword = env_variables["ADAFRUIT_IO_USERNAME"]
  defaultKey = env_variables["ADAFRUIT_IO_KEY"]
except KeyError as exc:
  print(f"Couldn't load password and username for Adafruit IO defaults: {exc}")
else:
  if defaultPassword and defaultKey:
    aio = Aio(username=defaultPassword, key=defaultKey)
    print("AIO: Loaded credentials from environment variables")


from functools import cache
from typing import Callable
from .do_test import test_host_import
host = test_host_import()

@cache
def test_host_main_exists():
  assert host.main
  return host.main

def test_host_funcs_exist():
  assert host.main
  assert host.step

def test_host_internal_funcs_exist():
  assert host._load
  assert host._print
  assert host._ping

def test_host_ping_works():
  assert host._ping
  host._ping(inactive=True, receiving=False)




# Original Source: https://docs.circuitpython.org/projects/mlx90640/en/latest/

try:
  import board
except NotImplementedError as exc:
  print(f"Note: This machine is not compatible with MLX90640")
else:
  import busio
  import adafruit_mlx90640

  i2c = busio.I2C(board.SCL, board.SDA, frequency=800000)

  mlx = adafruit_mlx90640.MLX90640(i2c)
  print("MLX addr detected on I2C", [hex(i) for i in mlx.serial_number])

  # if using higher refresh rates yields a 'too many retries' exception,
  # try decreasing this value to work with certain pi/camera combinations
  mlx.refresh_rate = adafruit_mlx90640.RefreshRate.REFRESH_2_HZ

def main():
  try:
    import Process_raw as process
  except ImportError as exc:
    print("Error: Could not import Process_raw, not started main loop")
  while True:
    process.iterate(process.print_frame_value)

def get_frame():
  frame = [0] * (24*32)
  try:
      mlx.getFrame(frame)
  except ValueError:
      # these happen, no biggie - retry
      print("(Error)")
      return None
  return frame

if __name__ == "__main__":
  main()


import json
from typing import Callable, Tuple, TypeVar
from .lib.AIO import aio
from .lib.ThermalSensor.Extract_raw import get_frame

T_Data = TypeVar('T_Data')
def _step(
  method: Callable[[], Tuple[bool, T_Data]],
  callback: Callable[[T_Data], Tuple[bool]],
) -> bool:
  status, data = method()
  if status is False: return False
  finished_successfully = callback(data)
  if not finished_successfully: return False
  return True

def _method():
  frame = get_frame()
  if frame is None:
    return False
  return True, get_frame

def _send(data):
  try:
    aio.send_stream_data(data)
    return True
  except json.JSONDecodeError:
    print('JSONDecodeError: Cannot jsonify data :(')
    aio.pi_status_send_code("JSONDecodeError - pi.py")
    return False

def step():
  aio.pi_ping()
  # Step for uploading data
  _step(
    method = _method,
    callback = _send,
  )
  # Other steps could be added below, such as 8x8 LED matrix

def main():
  print("Beginning pi ...")
  while True: 
    step()

if __name__ == "__main__":
  main()



import json
from typing import List
from .lib.AIO import aio
from .lib.ThermalSensor.Process_raw import print_frame

def _load(strict=True):
  data = None
  try:
    data = aio.receive_stream_data()
  except json.JSONDecodeError as exc:
    print(f"(Error: Received stream bad: {exc}")
    print(f"{loaded=}")
    aio.host_status_send("JSONDecodeError - host.py")
    if strict:
      raise TypeError(f"(Error: JSONDecodeError - host.py :: Strict=True)")
  return data

def _load_stream(stream):
  """Takes _load() and returns the parsed stream"""
  if type(stream) is not list:
    try:
      stream = json.loads(stream)
    except json.JSONDecodeError as exc:
      print("(Error while loading stream {exc=})")
      raise TypeError(f"Expected type list, got stream {stream=}")
  if len(stream) == 0:
    raise TypeError(f"Expected length > 0, got {len(stream)} for {stream=}")
  if max(stream) - min(stream) == 0:
    raise TypeError(f"Expected range > 0, got {max(stream)} - {min(stream)} for {stream=}")
  return stream

def _print(stream: List[int]):
  print_frame(stream)

def _ping(*x, **y):
  aio.host_ping(*x, **y)

def step() -> bool:
  _ping()
  try:
    raw = _load(strict=True)
    data = _load_stream(raw)
  except TypeError as exc:
    print(f"(Error while gathering stream (host.py): {exc})")
    return False
  if data is True:
    print("(Error: No data received)")
    return False
  _print(data)
  return True

def main():
  print("Beginning host ...")
  while True:
    step()

if __name__ == "__main__":
  main()


from functools import cache
import uuid


@cache
def test_pi_import():
  from . import pi
  assert pi
  return pi

@cache
def test_pi_main_exists():
  pi = test_pi_import()
  assert pi.main
  return pi.main

@cache
def test_host_import():
  from . import host
  assert host
  return host

@cache
def test_host_main_exists():
  host = test_host_import()
  assert host.main
  return host.main

@cache
def test_aio_lib_import():
  from .lib import AIO
  if AIO.aio is None:
    raise NotImplementedError("Please instinate AIO.aio with credentials (.env file)")
  assert AIO.aio.send_stream_data
  assert AIO.aio.receive_stream_data
  return AIO.aio

def check_AIO(post, receive, extractID=lambda s: s.split(":")[1]):
  id = str(uuid.uuid4())
  
  post(id)
  received = receive()
  
  assert received
  receivedID = extractID(received)
  assert receivedID == id

def test_aio_stream_works():
  aio = test_aio_lib_import()
  check_AIO(
    lambda id: aio.send_stream_data(f"Send from pytest (__stream_data__) test_aio_stream_works:{id}"),
    lambda: aio.receive_stream_data(),
  )
  return aio


# SPDX-FileCopyrightText: 2021 ladyada for Adafruit Industries
# SPDX-License-Identifier: MIT
# https://github.com/adafruit/Adafruit_CircuitPython_MLX90640/tree/main/examples

import time
import board
import busio
import adafruit_mlx90640

PRINT_TEMPERATURES = False
PRINT_ASCIIART = True

i2c = busio.I2C(board.SCL, board.SDA, frequency=800000)

mlx = adafruit_mlx90640.MLX90640(i2c)
print("MLX addr detected on I2C")
print([hex(i) for i in mlx.serial_number])

mlx.refresh_rate = adafruit_mlx90640.RefreshRate.REFRESH_2_HZ

frame = [0] * 768
while True:
    stamp = time.monotonic()
    try:
        mlx.getFrame(frame)
    except ValueError:
        # these happen, no biggie - retry
        continue
    print("Read 2 frames in %0.2f s" % (time.monotonic() - stamp))
    for h in range(24):
        for w in range(32):
            t = frame[h * 32 + w]
            if PRINT_TEMPERATURES:
                print("%0.1f, " % t, end="")
            if PRINT_ASCIIART:
                c = "&"
                # pylint: disable=multiple-statements
                if t < 20:
                    c = " "
                elif t < 23:
                    c = "."
                elif t < 25:
                    c = "-"
                elif t < 27:
                    c = "*"
                elif t < 29:
                    c = "+"
                elif t < 31:
                    c = "x"
                elif t < 33:
                    c = "%"
                elif t < 35:
                    c = "#"
                elif t < 37:
                    c = "X"
                # pylint: enable=multiple-statements
                print(c, end="")
        print()
    print()


from functools import cache
from typing import Any, Callable, List, Tuple
import uuid
from .do_test import test_pi_import
pi = test_pi_import()

@cache
def test_pi_main_exists():
  assert pi.main
  return pi.main

def test_pi_funcs_exists():
  assert pi.main
  assert pi.step

def test_pi_internel_funcs_exists():
  assert pi._step
  assert pi._method
  assert pi._step
  
def test_pi_internal_step():
  assert pi._step
  # Check if method and callback are successful
  def method_callback(return_success: bool = True, **kwargs):
    def _method_callback(data: str, check_called: List[bool | None]) -> Callable[[], Tuple[bool, str]]:
      def __method_callback() -> Tuple[bool, str]:
        check_called[0] = True
        return return_success, data
      return __method_callback
    return _method_callback
  
  def callback_callback(
    *,
    return_success: bool = True,
    data_should_be_same: bool | None = None,
    **kwargs
  ):
    def _callback_callback(should_data: str, check_called: List[bool | None] = [None, None]) -> Callable[[str], bool]:
      def __callback_callback(passed_data: Any) -> bool:
        check_called[1] = True
        if data_should_be_same is True:
          assert should_data == passed_data
        elif data_should_be_same is False:
          assert should_data != passed_data
        return return_success
      return __callback_callback
    return _callback_callback

  def check_step(
    method_gen: Callable[[str, List[bool | None]], Callable[[], bool]],
    callback_gen: Callable[[str, List[bool | None]], Callable[[str], bool]],
    callback_states: Tuple[bool | None, bool | None] = (None, None),
  ):
    check_called: List[bool | None] = [False, False]
    id = str(uuid.uuid4())
    pi._step(
      method_gen(id, check_called),
      callback_gen(id, check_called),
    )
    for state_num, state in enumerate(callback_states):
      if state is not None:
        assert check_called[state_num] is state
  
  check_step(
    method_callback(return_success=True),
    callback_callback(return_success=True, data_should_be_same=True),
    callback_states=(True, True),
  )
  check_step(
    method_callback(return_success=True),
    callback_callback(return_success=False, data_should_be_same=True),
    callback_states=(True, True),
  )
  check_step(
    method_callback(return_success=False),
    callback_callback(return_success=True, data_should_be_same=True),
    callback_states=(True, False),
  )
  check_step(
    method_callback(return_success=False),
    callback_callback(return_success=False, data_should_be_same=True),
    callback_states=(True, False),
  )  

# Test pings
# test _hidden funcs :)


try:
    import pretty_traceback
    pretty_traceback.install()
    print("Pretty traceback installed")
except ImportError:
    print("(Warning: pretty_traceback not installed)")
    pass    # no need to fail because of missing dev dependency


from functools import cache
import sys
from typing import Callable
import pytest
from rich import print as rprint

@cache
def test_process_import():
  from . import Process_raw as raw
  assert raw
  return raw

@pytest.fixture
def raw():
  return test_process_import()

def test_process_funcs_exists(raw):
  assert raw.iterate_frame
  assert raw.iterate
  assert raw.print_frame

def test_print_frame_value(raw, capfd):
    _list = [69, 42]
    raw.print_frame_value(_list[0], 0, 0, list=_list, dimensions=(2,1))
    capture1 = capfd.readouterr().out
    _avg = sum(_list) / len(_list)
    _colour = raw.generate_colour(69, avg=_avg)
    rprint(f"[{_colour}]{raw.ascii_chars[len(raw.ascii_chars)-1]}[/{_colour}]", end="")
    capture2 = capfd.readouterr().out
    assert capture1 == capture2
    
  



import json
from typing import TYPE_CHECKING, Any, Callable, Tuple
from functools import cache
import uuid
import pytest

if TYPE_CHECKING:
  from .AIO import Aio as T_Aio

@cache
def test_aio_import():
  from . import AIO
  return AIO

@cache
def test_aio_class_exists():
  aio = test_aio_import()
  assert aio.Aio
  return aio.Aio

@cache
def test_aio_has_instance():
  aio = test_aio_import()
  assert aio.aio is None or isinstance(aio.aio, aio.Aio)
  return aio.aio if aio.aio is not None else False

@cache
def test_env_credentials():
  from dotenv import dotenv_values
  secrets = dotenv_values(verbose=True)
  # print(f"Secrets: {secrets}")
  if secrets == None:
    return False
  # print("Checking for ADAFRUIT_IO_USERNAME in dotenv values ...")
  assert "ADAFRUIT_IO_USERNAME" in secrets
  # print("Checking for ADAFRUIT_IO_KEY in dotenv values ...")
  assert "ADAFRUIT_IO_KEY" in secrets
  return secrets

@cache
def test_proper_credentials():
  secrets = test_env_credentials()
  if secrets is None:
    return False
  aio = test_aio_has_instance()
  assert aio is not False
  assert aio.client
  assert aio.client.username == secrets["ADAFRUIT_IO_USERNAME"]
  assert aio.client.key == secrets["ADAFRUIT_IO_KEY"]
  return True, aio

@pytest.fixture
def aio():
  """Fixture for aio sign in"""
  status, _aio = test_proper_credentials()
  if status is False:
    raise NotImplementedError("No signed-in AIO instance")
  return _aio

@cache
def test_aio_instance_schema():
  AIO = test_aio_import()
  aio = test_aio_has_instance()
  if aio is False: return False
  assert aio.schema == AIO.defaultSchema
  assert aio.schema["group"]
  return True

def check_AIO(post, receive, extractID=lambda s: s.split(":")[1]):
  id = str(uuid.uuid4())
  
  post(id)
  received = receive()
  
  assert received
  receivedID = extractID(received)
  assert receivedID == id

def test_schema_send_receive(aio):
  check_AIO(
    lambda id: aio.send_schema("test", data=f"Send from pytest (schema test) test_schema_send_receive:{id}"),
    lambda: aio.receive_schema("test"),
  )
  for schema in aio.schema:
    if schema == "group": continue
    check_AIO(
      lambda id: aio.send_schema(schema, data=f"Send from pytest ([Auto] {schema}) test_schema_send_receive:{id}"),
      lambda: aio.receive_schema(schema),
    )

def test_raw_send_receive(aio):
  check_AIO(
    lambda id: aio.client.send("default.test", f"Send from pytest (default.test) test_raw_send_receive:{id}"),
    lambda: aio.client.receive("default.test").value
  )
  check_AIO(
    lambda id: aio.client.send("brad.test", f"Send from pytest (brad.test) test_raw_send_receive:{id}"),
    lambda: aio.client.receive("brad.test").value
  )

def test_custom_send_receive(aio):
  check_AIO(
    lambda id:
    aio.send("default", "test", data=f"Send from pytest (default.test) test_custom_send_receive:{id}"),
    lambda: aio.receive("default", "test")
  )
  check_AIO(
    lambda id:
    aio.send("brad", "test", data=f"Send from pytest (brad.test) test_custom_send_receive:{id}"),
    lambda: aio.receive("brad", "test")
  )

def test_schema_stream(aio):
  check_AIO(
    lambda id: aio.send_stream(data=json.dumps({
      "stream": [],
      "test": f"Send from pytest (schema stream) test_schema_stream:{id}",
    })),
    lambda: aio.receive_stream(),
    lambda s: json.loads(s)["test"].split(":")[1]
  )

def test_schema_stream_data(aio):
  check_AIO(
    lambda id: aio.send_stream_data(f"Send from pytest (schema stream data) test_schema_stream_data:{id}"),
    lambda: aio.receive_stream_data(),
  )

def test_pi_ping(aio):
  # Manually check ping data
  aio.pi_ping(inactive=True)
  assert int(aio._get_pi_ping_status()) == -1

def test_host_ping(aio):
  # Manually check ping data
  aio.host_ping(inactive=True)
  assert int(aio._get_host_ping_status()) == -1

def test_pi_status_updates(aio):
  check_AIO(
    lambda id: aio.pi_status_send(data=f"Send from pytest (pi status update) test_pi_status_updates:{id}"),
    lambda: aio._get_pi_status(),
  )

def test_host_status_updates(aio):
  check_AIO(
    lambda id: aio.host_status_send(data=f"Send from pytest (host status update) test_host_status_updates:{id}"),
    lambda: aio._get_host_status(),
  )

def test_reset_stream(aio):
  



from functools import cache
from typing import Callable

@cache
def test_extract_import():
  from . import Extract_raw as extract
  assert extract
  return extract

extract = test_extract_import()

def require_extract(f: Callable) -> Callable:
  """
  Decorator to ensure import of module before calling function.
  """
  def wrapper(*args, **kwargs):
    extract = test_extract_import()
    return f(*args, extract=extract, **kwargs)
  return wrapper

@require_extract
def test_extract_funcs_exist(extract):
  assert extract.get_frame
  assert extract.main
  


''' from .  '''
from typing import List, Tuple
from . import Extract_raw as raw
from .Extract_raw import get_frame
from rich import pretty as p
from rich import print as rprint
p.install()

ascii_chars = "$@B%8&WM#*oahkbdpqwmZO0QLCJUYXzcvunxrjft/\\|()1{}[]?-_+~<>i!lI;:,\"^`'."

lowest_temp = 15
highest_temp = 45

defaultDimensions = (32, 24)

def map_num_ranges(value, leftMin, leftMax, rightMin=0, rightMax=int(len(ascii_chars) - 1)):
    # Figure out how 'wide' each range is
    leftSpan = leftMax - leftMin
    rightSpan = rightMax - rightMin

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(value - leftMin) / float(leftSpan)

    # Convert the 0-1 range into a value in the right range.
    v = int(rightMin + (valueScaled * rightSpan))
    return max(min(v, rightMax), rightMin)

def get_ascii_char_from_num(num, *, min=lowest_temp, max=highest_temp):
  """Put in number from 15-45 and get back ascii char from ascii_chars"""
  if num < min:
    num = min
  if num > max:
    num = max
  charNum = map_num_ranges(num, min, max)
  # print(f"{charNum=}")
  return ascii_chars[charNum]

def generate_colour(temp, *, avg):
  return "red" if temp > avg else "green"

def get_char(temp, *, list, previous=None, final=False):
  assert len(list) > 0 # Div by 0 when calculating average
  assert min(list) - max(list) != 0 # Div by 0 when calculating / range
  
  avg = sum(list) / len(list)
  colour = generate_colour(temp, avg=avg)
  beginningColours = f"[/{previous}][{colour}]" if previous is not colour else ""
  beginningColours = f"[{colour}]" if previous is None else beginningColours
  endingColours = f"[/{colour}]" if final else ""
  chosenChar = get_ascii_char_from_num(int(temp), min=min(list), max=max(list))
  return f"{beginningColours}{chosenChar}{endingColours}"

def print_frame_value(value, x, y, list, *, previous=None, dimensions=defaultDimensions, final=False):
  assert value in list
  final = final or bool(x == dimensions[0]-1 and y == dimensions[1]-1)
  toPrint = get_char(value, list=list, previous=previous, final=final)
  rprint(toPrint, end="")
  if x == dimensions[0]-1:
    print() # newline every line of chars
    if y == dimensions[1]-1:
      print() # newline at end of frame

def iterate_frame(frame, *, dimensions=defaultDimensions):
  for y in range(dimensions[1]):
    for x in range(dimensions[0]):
      yield frame[y*dimensions[0] + x], x, y

def iterate(f, frame:List[int]|None=None, *, dimensions:Tuple[int, int]|None=defaultDimensions):
  if frame is None:
    frame = get_frame()
    if frame is None: return frame # retry, return None
  if dimensions is None and len(frame) == defaultDimensions[0] * defaultDimensions[1]:
    dimensions = defaultDimensions
  for value, x, y in iterate_frame(frame, dimensions=dimensions):
    f(value, x, y, frame, dimensions=dimensions)

def print_frame(frame=None):
  iterate(print_frame_value, frame=frame)